package safegoroutines_test

import (
	"context"
	"io/ioutil"
	"log"
	"sync"
	"testing"
	"time"
)

func init() {
	log.SetOutput(ioutil.Discard)
}

// この実装方法はあくまで簡単にサンプルコードを記す
// ためのものであり、実際のコードで使うのはおすすめ
// しません。
type doneFlag struct {
	ch chan struct{}
	mu sync.Mutex
}

func (d *doneFlag) Done() chan struct{} {
	d.mu.Lock()
	defer d.mu.Unlock()

	if d.ch == nil {
		d.ch = make(chan struct{})
	}

	return d.ch
}

type ExportedField struct {
	doneFlag
	Listen string
}

func (s *ExportedField) Run() {
	defer close(s.Done())

	timeout := time.NewTimer(3 * time.Second)
	defer timeout.Stop()

	tick := time.NewTicker(10 * time.Millisecond)
	for {
		select {
		case <-tick.C:
			log.Print(s.Listen) // 利用する
		case <-timeout.C:
			return // 終了
		}
	}
}

// 内部情報を格納するフィールドが公開されている場合のテスト
func TestExportedField(t *testing.T) {
	var s ExportedField
	s.Listen = "127.0.0.1:9999"

	go s.Run()

	// 以下のループは本来必要のないものだが、レースコンディションを
	// 発生しやすくするために記述してある
	for {
		select {
		case <-s.Done():
			return
		default:
			s.Listen = "127.0.0.1:8888"
		}
	}
}

type ExportedMutatorMethod struct {
	doneFlag
	listen string
}

func NewExportedMutatorMethod(addr string) *ExportedMutatorMethod {
	return &ExportedMutatorMethod{
		listen: addr,
	}
}

func (s *ExportedMutatorMethod) SetDefaultListenAddress() {
	s.listen = "127.0.0.1:8000"
}

func (s *ExportedMutatorMethod) Run() {
	defer close(s.Done())

	timeout := time.NewTimer(3 * time.Second)
	defer timeout.Stop()

	tick := time.NewTicker(10 * time.Millisecond)
	for {
		select {
		case <-tick.C:
			log.Print(s.listen) // 利用する
		case <-timeout.C:
			return // 終了
		}
	}
}

// 内部情報を変更するメソッドが公開されている場合のテスト
func TestExportedMutatorMethod(t *testing.T) {
	var s = NewExportedMutatorMethod("127.0.0.1:9999")
	go s.Run()

	// 以下のループは本来必要のないものだが、レースコンディションを
	// 発生しやすくするために記述してある
	for {
		select {
		case <-s.Done():
			return
		default:
			s.SetDefaultListenAddress()
		}
	}
}

type ClosedState struct {
	doneFlag
	listen string
	mu     sync.Mutex
}

type closedStateRunner struct {
	listen string
}

func NewClosedState(addr string) *ClosedState {
	return &ClosedState{
		listen: addr,
	}
}

// このメソッドは本来必要ないが、ClosedStateの内部情報を
// 変更してもclosedStateRunnerとは関係がないことをテスト内で
// 確認するために実装してある
func (s *ClosedState) SetListenAddress(addr string) {
	s.mu.Lock()
	defer s.mu.Unlock()
	s.listen = addr
}

func (s *ClosedState) Run() {
	var r closedStateRunner

	s.mu.Lock()
	r.listen = s.listen
	s.mu.Unlock()

	defer close(s.Done())

	r.Run()
}

func (r *closedStateRunner) Run() {
	timeout := time.NewTimer(3 * time.Second)
	defer timeout.Stop()

	tick := time.NewTicker(10 * time.Millisecond)
	tick.Stop()
	for {
		select {
		case <-tick.C:
			log.Print(r.listen) // 利用する
		case <-timeout.C:
			return // 終了
		}
	}
}

// 状態を「閉じた」場合のテスト
func TestClosedState(t *testing.T) {
	var s = NewClosedState("127.0.0.1:9999")
	go s.Run()

	// 以下のループは本来必要のないものだが、レースコンディションを
	// 発生しやすくするために記述してある
	for {
		select {
		case <-s.Done():
			return
		default:
			s.SetListenAddress("127.0.0.1:8080")
		}
	}
}

type ClosedStateWithContext struct {
	listen string
	mu     sync.Mutex
}

type closedStateRunnerWithContext struct {
	listen string
}

func NewClosedStateWithContext(addr string) *ClosedStateWithContext {
	return &ClosedStateWithContext{
		listen: addr,
	}
}

// このメソッドは本来必要ないが、ClosedStateの内部情報を
// 変更してもclosedStateRunnerとは関係がないことをテスト内で
// 確認するために実装してある
func (s *ClosedStateWithContext) SetListenAddress(addr string) {
	s.mu.Lock()
	defer s.mu.Unlock()
	s.listen = addr
}

func (s *ClosedStateWithContext) Run(ctx context.Context) {
	var r closedStateRunnerWithContext

	s.mu.Lock()
	r.listen = s.listen
	s.mu.Unlock()

	r.Run(ctx)
}

func (r *closedStateRunnerWithContext) Run(ctx context.Context) {
	tick := time.NewTicker(10 * time.Millisecond)
	defer tick.Stop()
	for {
		select {
		case <-tick.C:
			log.Print(r.listen) // 利用する
		case <-ctx.Done():
			return // 終了
		}
	}
}

func TestClosedStateWithContext(t *testing.T) {
	ctx, cancel := context.WithTimeout(context.Background(), 3*time.Second)
	defer cancel()

	var s = NewClosedStateWithContext("127.0.0.1:9999")
	go s.Run(ctx)

	// 以下のループは本来必要のないものだが、レースコンディションを
	// 発生しやすくするために記述してある
	for {
		select {
		case <-ctx.Done():
			return
		default:
			s.SetListenAddress("127.0.0.1:8080")
		}
	}
}
